'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function createPool(provider, options) {
    var _a, _b, _c, _d;
    const poolOptions = (_a = options.poolOptions) !== null && _a !== void 0 ? _a : {};
    const connectionTtl = (_b = poolOptions.connectionTtl) !== null && _b !== void 0 ? _b : 60;
    const logMonitoring = (_c = poolOptions.logMonitoring) !== null && _c !== void 0 ? _c : (() => {
        // Nothing to do.
    });
    const keepOneConnection = !!poolOptions.keepOneConnection;
    // eslint-disable-next-line no-console
    const logError = (_d = options.logError) !== null && _d !== void 0 ? _d : (err => console.error(err));
    const logDebug = options.logDebug;
    if (logDebug)
        provider = debugWrapProvider(provider, logDebug);
    let closed = false;
    let available = [];
    let nonExclusiveDb;
    let nonExclusiveCount = 0;
    let cleaning;
    let counter = 0;
    const identifiers = new WeakMap();
    return {
        grab: async (exclusive = false) => {
            if (closed)
                throw new Error("Invalid call to \"grab\", the pool is closed");
            if (!exclusive && nonExclusiveDb) {
                ++nonExclusiveCount;
                return nonExclusiveDb;
            }
            const item = available.pop();
            let db;
            if (item)
                db = item.db;
            else {
                db = await provider();
                const id = ++counter;
                if (logDebug)
                    db = debugWrapAsyncMethods(db, id, logDebug);
                identifiers.set(db, id);
                logMonitoring({ event: "open", cn: db, id: identifiers.get(db) });
            }
            if (!exclusive) {
                ++nonExclusiveCount;
                nonExclusiveDb = db;
            }
            logMonitoring({ event: "grab", cn: db, id: identifiers.get(db) || -123 });
            return db;
        },
        release: (db) => {
            logMonitoring({ event: "release", cn: db, id: identifiers.get(db) });
            if (db === nonExclusiveDb) {
                if (--nonExclusiveCount === 0) {
                    nonExclusiveDb = undefined;
                    available.push({ db, releaseTime: Date.now() });
                }
            }
            else
                available.push({ db, releaseTime: Date.now() });
            if (closed)
                cleanOldConnections(true);
            else
                startCleaning();
        },
        abandon: (db) => {
            logMonitoring({ event: "abandon", cn: db, id: identifiers.get(db) });
            if (db === nonExclusiveDb) {
                --nonExclusiveCount;
                nonExclusiveDb = undefined;
            }
            db.close().catch(err => logError(err));
        },
        close: async () => {
            if (closed)
                throw new Error("Invalid call to \"close\", the pool is already closed");
            closed = true;
            const closeAll = available.map(item => {
                logMonitoring({ event: "close", cn: item.db, id: identifiers.get(item.db) });
                return item.db.close();
            });
            available = [];
            await Promise.all(closeAll);
            if (cleaning)
                clearInterval(cleaning);
        }
    };
    function startCleaning() {
        if (cleaning)
            return;
        let stopCount = 0;
        cleaning = setInterval(() => {
            cleanOldConnections();
            if (available.length > 0)
                stopCount = 0;
            else if (++stopCount >= 10) {
                clearInterval(cleaning);
                cleaning = undefined;
            }
        }, 1000);
        cleaning.unref();
    }
    function cleanOldConnections(force = false) {
        const olderThanTime = Date.now() - 1000 * connectionTtl;
        let index;
        const lastIndex = available.length - 1;
        for (index = 0; index <= lastIndex; ++index) {
            if (!force) {
                const tooOld = available[index].releaseTime > olderThanTime;
                const keepThisOne = keepOneConnection && !nonExclusiveDb && index === lastIndex;
                if (tooOld || keepThisOne)
                    break;
            }
            const db = available[index].db;
            logMonitoring({ event: "close", cn: db, id: identifiers.get(db) });
            db.close().catch(err => logError(err));
        }
        if (index > 0)
            available = available.slice(index);
    }
    function debugWrapProvider(provider, logDebug) {
        return async () => {
            try {
                const result = await provider();
                logDebug({ result });
                return result;
            }
            catch (error) {
                logDebug({ error });
                throw error;
            }
        };
    }
    function debugWrapAsyncMethods(db, idInPool, logDebug) {
        const inTransactions = new WeakSet();
        const wrap = {};
        for (const name of Object.keys(db)) {
            wrap[name] = async (...args) => {
                if (name === "exec" && args.length === 1 && args[0] === "begin")
                    inTransactions.add(wrap);
                try {
                    const result = await db[name](...args);
                    logDebug({
                        callingContext: {
                            connection: db,
                            method: name,
                            args,
                            inTransaction: inTransactions.has(wrap),
                            idInPool,
                        },
                        result
                    });
                    return result;
                }
                catch (error) {
                    logDebug({
                        callingContext: {
                            connection: db,
                            method: name,
                            args,
                            inTransaction: inTransactions.has(wrap),
                            idInPool
                        },
                        error
                    });
                    throw error;
                }
            };
        }
        return wrap;
    }
}

function toSingleRow(rows) {
    if (rows.length !== 1) {
        if (rows.length === 0)
            return;
        throw new Error(`Cannot fetch one row, row count: ${rows.length}`);
    }
    return rows[0];
}
function toSingleValue(row) {
    if (row === undefined)
        return;
    const columns = Object.keys(row);
    if (columns.length !== 1)
        throw new Error(`Cannot fetch one value, column count: ${columns.length}`);
    return row[columns[0]];
}
function formatError(err) {
    if (err instanceof Error)
        return err;
    return new Error(err);
}

class CursorProvider {
    constructor(context) {
        this.context = context;
        this.items = new Set();
    }
    async open(sql, params) {
        this.context.check.parameters(params);
        const { pool } = this.context;
        const cn = await pool.grab();
        try {
            const inst = new CursorItem({
                context: this.context,
                end: (item) => {
                    this.items.delete(item);
                    pool.release(cn);
                }
            }, await cn.cursor(sql, params));
            this.items.add(inst);
            return inst.cursor;
        }
        catch (err) {
            throw formatError(err);
        }
    }
    async closeAll() {
        await Promise.all(Array.from(this.items).map(item => item.close()));
    }
}
class CursorItem {
    constructor(itemContext, ac) {
        this.cursor = this.toCursor(itemContext, ac);
    }
    async close() {
        if (this.cursor.return)
            await this.cursor.return();
    }
    toCursor(itemContext, ac) {
        const obj = {
            [Symbol.asyncIterator]: () => obj,
            next: async () => {
                if (!ac)
                    return { done: true, value: undefined };
                const result = await ac.next();
                if (result.done) {
                    ac = undefined;
                    itemContext.end(this);
                }
                return result;
            },
            return: async () => {
                if (!ac)
                    return { done: true, value: undefined };
                itemContext.end(this);
                const returnCb = ac.return;
                ac = undefined;
                if (returnCb)
                    return await returnCb();
                return { done: true, value: undefined };
            },
            throw: async (err) => {
                if (!ac)
                    throw err;
                itemContext.end(this);
                const throwCb = ac.return;
                ac = undefined;
                if (throwCb)
                    await throwCb(err);
                throw err;
            }
        };
        return obj;
    }
}
// function makeInMemoryACursor(rows?: any[]): AsyncIterableIterator<any> {
//   let currentIndex = -1
//   const obj: AsyncIterableIterator<any> = {
//     [Symbol.asyncIterator]: () => obj,
//     next: async () => {
//       if (!rows)
//         return { done: true, value: undefined }
//       const value = rows[++currentIndex]
//       if (!value)
//         rows = undefined
//       return { done: !rows, value }
//     },
//     return: async () => {
//       rows = undefined
//       return { done: true, value: undefined }
//     },
//     throw: async err => {
//       rows = undefined
//       throw err
//     }
//   }
//   return obj
// }

function toExecResult(result) {
    const obj = {
        get affectedRows() {
            return result.affectedRows;
        },
        getInsertedId(options) {
            const id = result.getInsertedId(options);
            if (id === undefined)
                throw new Error("Missing inserted ID");
            return id;
        },
        getInsertedIdAsString(options) {
            const val = obj.getInsertedId(options);
            switch (typeof val) {
                case "string":
                    return val;
                case "number":
                    return val.toString();
                default:
                    throw new Error(`Unexpected inserted ID type: ${typeof val}`);
            }
        },
        getInsertedIdAsNumber(options) {
            const val = obj.getInsertedId(options);
            switch (typeof val) {
                case "string":
                    return parseInt(val, 10);
                case "number":
                    return val;
                default:
                    throw new Error(`Unexpected inserted ID type: ${typeof val}`);
            }
        }
    };
    return obj;
}

class PsProvider {
    constructor(ppContext) {
        this.ppContext = ppContext;
        this.items = new Set();
    }
    async prepare(sql) {
        const item = await PsItem.create(await this.createPsContext(), sql);
        this.items.add(item);
        return item.ps;
    }
    async closeAll() {
        await Promise.all(Array.from(this.items).map(item => item.ps.close()));
    }
    hasCursor() {
        for (const item of this.items) {
            if (item.hasCursor())
                return true;
        }
        return false;
    }
    async createPsContext() {
        const { exclusiveCn } = this.ppContext;
        if (exclusiveCn) {
            return {
                context: this.ppContext.context,
                cn: exclusiveCn,
                end: (item) => {
                    this.items.delete(item);
                },
                canCreateCursor: () => this.ppContext.canCreateCursor()
            };
        }
        const { pool } = this.ppContext.context;
        const cn = await pool.grab();
        return {
            context: this.ppContext.context,
            cn,
            end: (item) => {
                this.items.delete(item);
                pool.release(cn);
            },
            canCreateCursor: () => this.ppContext.canCreateCursor()
        };
    }
}
class PsItem {
    constructor(itemContext, aps) {
        this.ps = this.toPs(itemContext, aps);
    }
    static async create(psContext, sql) {
        const aps = await psContext.cn.prepare(sql);
        return new PsItem(psContext, aps);
    }
    hasCursor() {
        return !!this.cursorItem;
    }
    toPs(itemContext, aps) {
        let obj = {
            exec: async (params) => {
                itemContext.context.check.parameters(params);
                check("exec");
                try {
                    return toExecResult(await aps.exec(mergeParameters(this.boundParams, params)));
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            all: async (params) => {
                itemContext.context.check.parameters(params);
                check("all");
                try {
                    return await aps.all(mergeParameters(this.boundParams, params));
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            singleRow: async (params) => toSingleRow(await obj.all(params)),
            singleValue: async (params) => toSingleValue(await obj.singleRow(params)),
            cursor: async (params) => {
                itemContext.context.check.parameters(params);
                check("cursor");
                if (!itemContext.context.capabilities.cursors)
                    throw new Error("Cursors are not available with this adapter");
                if (!itemContext.canCreateCursor())
                    throw new Error("Only one cursor is allowed by underlying transaction");
                try {
                    const aCursor = await aps.cursor(mergeParameters(this.boundParams, params));
                    this.cursorItem = new CursorItem({
                        context: itemContext.context,
                        end: () => {
                            this.cursorItem = undefined;
                        }
                    }, aCursor);
                    return this.cursorItem.cursor;
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            bind: async (paramsOrIndexOrKey, value) => {
                check("bind");
                if (typeof paramsOrIndexOrKey === "object") {
                    itemContext.context.check.parameters(paramsOrIndexOrKey);
                    this.boundParams = mergeParameters(this.boundParams, paramsOrIndexOrKey);
                }
                else {
                    if (typeof paramsOrIndexOrKey === "string") {
                        itemContext.context.check.namedParameters();
                    }
                    this.boundParams[paramsOrIndexOrKey] = value;
                }
                return Promise.resolve();
            },
            unbind: async (indexOrKey) => {
                check("unbind");
                if (indexOrKey === undefined)
                    this.boundParams = undefined;
                else {
                    if (typeof indexOrKey === "string") {
                        itemContext.context.check.namedParameters();
                    }
                    this.boundParams[indexOrKey] = undefined;
                }
                return Promise.resolve();
            },
            close: async () => {
                if (!aps)
                    throw new Error("Prepared statement is already closed");
                const copy = aps;
                aps = undefined;
                if (this.cursorItem)
                    await this.cursorItem.close();
                try {
                    await copy.close();
                }
                catch (err) {
                    throw formatError(err);
                }
                itemContext.end(this);
            }
        };
        if (itemContext.context.options.modifier && itemContext.context.options.modifier.modifyPreparedStatement)
            obj = itemContext.context.options.modifier.modifyPreparedStatement(obj);
        return obj;
        function check(method) {
            if (!aps)
                throw new Error(`Invalid call to '${method}', the prepared statement is closed`);
        }
    }
}
function mergeParameters(params1, params2) {
    if (!params1)
        return params2;
    if (!params2)
        return params1;
    const isArr = Array.isArray(params1);
    if (isArr !== Array.isArray(params2))
        throw new Error("Cannot merge named parameters with positioned parameters");
    if (isArr) {
        const result = [...params1];
        const p2 = params2;
        p2.forEach((val, index) => result[index] = val);
        return result;
    }
    return {
        ...params1,
        ...params2
    };
}

class TxProvider {
    constructor(context) {
        this.context = context;
        this.items = new Set();
    }
    async create() {
        const item = await TxItem.create({
            context: this.context,
            end: (item) => {
                this.items.delete(item);
            }
        });
        this.items.add(item);
        return item.tx;
    }
    async closeAll() {
        await Promise.all(Array.from(this.items).map(item => item.tx.rollback()));
    }
}
class TxItem {
    constructor(itemContext, acn) {
        this.tx = this.toTx(itemContext, acn);
    }
    static async create(txContext) {
        const acn = await txContext.context.pool.grab(true);
        try {
            if (txContext.context.hooks.beginTransaction)
                await txContext.context.hooks.beginTransaction(acn);
            else
                await acn.exec("begin");
            return new TxItem(txContext, acn);
        }
        catch (err) {
            throw formatError(err);
        }
    }
    hasCursor() {
        return !!this.cursorItem;
    }
    canCreateCursor() {
        return !this.cursorItem && (!this.psProvider || !this.psProvider.hasCursor());
    }
    async closeDependencies() {
        const promises = [];
        if (this.cursorItem)
            promises.push(this.cursorItem.close());
        if (this.psProvider)
            promises.push(this.psProvider.closeAll());
        await Promise.all(promises);
    }
    toTx(itemContext, acn) {
        const endOfTransaction = async (method, item) => {
            if (!acn)
                throw new Error(`Invalid call to '${method}', not in a transaction`);
            const copy = acn;
            acn = undefined;
            itemContext.end(item);
            try {
                await item.closeDependencies();
                const hook = itemContext.context.hooks[method];
                try {
                    if (hook)
                        await hook(copy);
                    else
                        await copy.exec(method);
                }
                catch (err) {
                    throw formatError(err);
                }
                itemContext.context.pool.release(copy);
            }
            catch (err) {
                itemContext.context.pool.abandon(copy);
                throw err;
            }
        };
        let obj = {
            prepare: async (sql) => {
                itemContext.context.check.preparedStatements();
                if (!acn)
                    throw new Error("Invalid call to 'prepare', the connection is closed");
                if (!this.psProvider) {
                    this.psProvider = new PsProvider({
                        context: itemContext.context,
                        exclusiveCn: acn,
                        canCreateCursor: () => this.canCreateCursor()
                    });
                }
                return await this.psProvider.prepare(sql);
            },
            async exec(sql, params) {
                itemContext.context.check.parameters(params);
                if (!acn)
                    throw new Error("Invalid call to 'exec', not in a transaction");
                try {
                    return toExecResult(await acn.exec(sql, params));
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            async all(sql, params) {
                itemContext.context.check.parameters(params);
                if (!acn)
                    throw new Error("Invalid call to 'all', not in a transaction");
                try {
                    return acn.all(sql, params);
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            singleRow: async (sql, params) => toSingleRow(await obj.all(sql, params)),
            singleValue: async (sql, params) => toSingleValue(await obj.singleRow(sql, params)),
            cursor: async (sql, params) => {
                itemContext.context.check.cursors();
                itemContext.context.check.parameters(params);
                if (!acn)
                    throw new Error("Invalid call to 'cursor', not in a transaction");
                if (!this.canCreateCursor())
                    throw new Error("Only one cursor is allowed by underlying transaction");
                try {
                    this.cursorItem = new CursorItem({
                        context: itemContext.context,
                        end: () => {
                            this.cursorItem = undefined;
                        }
                    }, await acn.cursor(sql, params));
                    return this.cursorItem.cursor;
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            async script(sql) {
                itemContext.context.check.script();
                if (itemContext.context.capabilities.script === "onASeparateConnection")
                    throw new Error("Scripts are available only on the main connection with this adapter.");
                if (!acn)
                    throw new Error("Invalid call to 'script', not in a transaction");
                try {
                    return acn.script(sql);
                }
                catch (err) {
                    throw formatError(err);
                }
            },
            get inTransaction() {
                return !!acn;
            },
            commit: () => endOfTransaction("commit", this),
            rollback: () => endOfTransaction("rollback", this)
        };
        if (itemContext.context.options.modifier && itemContext.context.options.modifier.modifyConnection)
            obj = itemContext.context.options.modifier.modifyConnection(obj);
        return obj;
    }
}

function makeMainConnection(context) {
    const { pool } = context;
    const psProvider = new PsProvider({
        context,
        canCreateCursor: () => true
    });
    const txProvider = new TxProvider(context);
    const cursorProvider = new CursorProvider(context);
    let closed = false;
    let obj = {
        async prepare(sql) {
            context.check.preparedStatements();
            if (closed)
                throw new Error("Invalid call to 'prepare', the connection is closed");
            return await psProvider.prepare(sql);
        },
        async exec(sql, params) {
            context.check.parameters(params);
            if (closed)
                throw new Error("Invalid call to 'exec', the connection is closed");
            const cn = await pool.grab();
            try {
                const res = await cn.exec(sql, params);
                return toExecResult(res);
            }
            finally {
                pool.release(cn);
            }
        },
        async all(sql, params) {
            context.check.parameters(params);
            if (closed)
                throw new Error("Invalid call to 'all', the connection is closed");
            const cn = await pool.grab();
            try {
                return await cn.all(sql, params);
            }
            catch (err) {
                throw formatError(err);
            }
            finally {
                pool.release(cn);
            }
        },
        async singleRow(sql, params) {
            return toSingleRow(await this.all(sql, params));
        },
        async singleValue(sql, params) {
            return toSingleValue(await this.singleRow(sql, params));
        },
        async cursor(sql, params) {
            context.check.cursors();
            context.check.parameters(params);
            if (closed)
                throw new Error("Invalid call to 'cursor', the connection is closed");
            return await cursorProvider.open(sql, params);
        },
        async script(sql) {
            context.check.script();
            if (closed)
                throw new Error("Invalid call to 'script', the connection is closed");
            if (context.capabilities.script === "onASeparateConnection") {
                const cn = await context.provider({ enableScript: true });
                try {
                    return await cn.script(sql);
                }
                catch (err) {
                    throw formatError(err);
                }
                finally {
                    await cn.close();
                }
            }
            else {
                const cn = await pool.grab();
                try {
                    return await cn.script(sql);
                }
                catch (err) {
                    throw formatError(err);
                }
                finally {
                    pool.release(cn);
                }
            }
        },
        beginTransaction: async () => {
            if (closed)
                throw new Error("Invalid call to 'beginTransaction', the connection is closed");
            return await txProvider.create();
        },
        close: async () => {
            if (closed)
                throw new Error("Invalid call to 'close', the connection is already closed");
            closed = true;
            await Promise.all([
                psProvider.closeAll(),
                cursorProvider.closeAll(),
                txProvider.closeAll()
            ]);
            await pool.close();
        }
    };
    if (context.options.modifier && context.options.modifier.modifyConnection)
        obj = context.options.modifier.modifyConnection(obj);
    return obj;
}

function ladc(options) {
    var _a;
    const provider = async (createOptions) => {
        const cn = await options.adapter.createConnection(createOptions);
        if (options.initConnection)
            await options.initConnection(cn);
        return cn;
    };
    return makeMainConnection({
        options,
        pool: createPool(provider, options),
        provider,
        capabilities: options.adapter.capabilities,
        hooks: (_a = options.adapter.hooks) !== null && _a !== void 0 ? _a : {},
        check: makeCheckers(options.adapter.capabilities)
    });
}
function makeCheckers(capabilities) {
    return {
        cursors() {
            if (!capabilities.cursors)
                throw new Error("Cursors are not available with this adapter.");
        },
        namedParameters() {
            if (!capabilities.namedParameters)
                throw new Error("Named parameters are not available with this adapter.");
        },
        preparedStatements() {
            if (!capabilities.preparedStatements)
                throw new Error("Prepared statements are not available with this adapter.");
        },
        script() {
            if (!capabilities.script)
                throw new Error("Scripts are not available with this adapter.");
        },
        parameters(params) {
            if (params && !Array.isArray(params) && !capabilities.namedParameters)
                throw new Error("Named parameters are not available with this adapter.");
        }
    };
}

exports.default = ladc;
